#ifndef STIM_ARGUMENTS
#define STIM_ARGUMENTS

#include <string>
#include <vector>
#include <iostream>
#include <iomanip>
#include <sstream>
#include <iterator>
#include <algorithm>

#ifdef _WIN32
#include <Windows.h>
#endif

namespace stim{

	class cmd_option
	{
	private:
		bool ansi;

		//argument name
		std::string name;

		//description of the argument
		std::vector<std::string> desc;

		//argument values
		std::vector<std::string> vals;

		//range or example
		std::string range;

		//flag is true when the argument is user-specified
		bool flag;

		void parse_val(const std::string &s){

			vals.clear();

			std::stringstream ss(s);
			std::string item;
			while (std::getline(ss, item, ' ')) {
				vals.push_back(item);
			}
		}

		void parse_desc(const std::string &s){

			desc.clear();

			std::stringstream ss(s);
			std::string item;
			while (std::getline(ss, item, '\n')) {
				desc.push_back(item);
			}
		}

	public:
		void set_ansi(bool b){ ansi = b; }
        //create an option with a given name, description, and default value
		cmd_option(std::string _name, std::string _desc, std::string _default = "", std::string _range = "")
		{
			name = _name;
			parse_desc(_desc);
			parse_val(_default);

			//if a default value is provided, set the flag
			if(_default != "")
                flag = true;
            else flag = false;

			range = _range;


		}

		int nargs()
		{
            return vals.size();
        }

		//return the value of a text option
		std::string as_string(unsigned int n = 0)
		{
            if(!flag)
            {
                std::cout<<"ERROR - Option requested without being set: "<<name<<std::endl;
                exit(1);
            }

            if(vals.size() > n)
                return vals[n];

            else return "";
		}

        //return the value of a floating point option
		float as_float(unsigned int n = 0)
		{
            if(!flag)
            {
                std::cout<<"ERROR - option requested without being set: "<<name<<std::endl;
                exit(1);
            }

            if(vals.size() > n)
            {
                float r;
                if ( ! (std::istringstream(vals[n]) >> r) ) r = 0;
                return r;
            }

            else return 0;
		}

		//return the value of an integer option
		int as_int(unsigned int n = 0)
		{
            if(!flag)
            {
                std::cout<<"ERROR - option requested without being set: "<<name<<std::endl;
                exit(1);
            }

            if(vals.size() > n)
            {
                int r;
                if ( ! (std::istringstream(vals[n]) >> r) ) r = 0;
                return r;
            }

            else return 0;
		}

		//get the width of the left column
		int col_width()
		{
            int n = 3;
            //add the length of the option name
            n += name.size();

            //if there are any default parameters
            if(vals.size() > 0)
            {
                //padding (parenthesis, =, etc.)
                n += 6;

                //for each default option value
                for(unsigned int v=0; v<vals.size(); v++)
                    n += vals[v].size() + 1;
            }

            //add a buffer of 4 characters
            n += 4;

            return n;
		}


		//string output
		std::string toStr(int width = 0)
		{
			std::stringstream ss;

			int color_size = 0;


			//create the left column
			std::string left_part = std::string(" --") + name;
			if(vals.size() != 0)
			{
				if(ansi)
					left_part += "\033[1;32m";
				left_part += " ( = ";
				for(unsigned int v=0; v<vals.size(); v++)
					left_part += vals[v] + std::string(" ");
				left_part += ")";
				if(ansi)
					left_part += "\033[0m";
				if(ansi)
					color_size = 11;
			}
			else
                color_size = 0;

			//if no width is passed, put 4 spaces between left and right columns
			if(width == 0) width = col_width();

			ss<<std::left<<std::setw(width + color_size)<<left_part;

			//output right column
			for(unsigned int d=0; d<desc.size(); d++)
			{
				if(d == 0)
					ss<<desc[0];
				else
					ss<<std::endl<<std::setfill(' ')<<std::setw(width)<<" "<<desc[d];

			}

			//output the range in the right column
			if(range != "" && ansi)
                ss<<std::endl<<std::setfill(' ')<<std::setw(width)<<" "<<"    "<<std::string("\033[1;36m") + range + "\033[0m";
			else if(range != "")
				ss<<std::endl<<std::setfill(' ')<<std::setw(width)<<" "<<"    "<<range;

			return ss.str();
		}

		//compare the name of the option to a string
		bool operator==(std::string rhs)
		{
            return (name == rhs);
		}

		//set the option to a given value
		void set(std::string _value)
		{
            parse_val(_value);

            //set the flag
            flag = true;
		}

		bool is_set()
		{
            return flag;
        }

	};

	struct argsection
	{
		std::string name;
		unsigned int index;
	};

	class arglist
	{
    private:
		bool ansi;

		//vector of options
        std::vector<cmd_option> opts;
        std::vector<std::string> args;

		//column width of the longest option
        int col_width;

		//list of sections
		std::vector<argsection> sections;

    public:

        arglist(){
        	col_width = 0;
        	ansi = true;
        }

		void set_ansi(bool b)
		{
			ansi = b;
			for(unsigned int i=0; i<opts.size(); i++)
				opts[i].set_ansi(ansi);
		}

        void add(std::string _name, std::string _desc, std::string _default = "", std::string _range = "")
        {
            cmd_option opt(_name, _desc, _default, _range);
			opt.set_ansi(ansi);
            opts.push_back(opt);

            col_width = std::max<int>(col_width, opt.col_width());
        }

		void section(std::string _name)
		{
			argsection s;
			s.name = _name;
			s.index = opts.size();
			sections.push_back(s);
		}

        //output the options (generally in response to --help)
        std::string str()
        {
            std::stringstream ss;

            int si = -1;

            if(sections.size() > 0)
                si = 0;

            //for each option
            for(unsigned int a=0; a<opts.size(); a++)
            {
                if(si != -1 && a == sections[si].index)
                {
					if(ansi)
						ss<<std::endl<<std::left<<std::setfill('=')<<std::setw(col_width)<<std::string("\033[1;31m") + sections[si].name<<"\033[0m"<<std::endl;
					else
						ss<<std::endl<<std::left<<std::setfill('=')<<std::setw(col_width)<<sections[si].name<<std::endl;
                    si++;
                    if(si == (int)sections.size()) si = -1;
                }

                ss<<opts[a].toStr(col_width)<<std::endl;
            }

            return ss.str();
        }

        int index(std::string _name)
        {
        	unsigned int i = find(opts.begin(), opts.end(), _name) - opts.begin();

            if(i >= opts.size())
                return -1;

            return (int)i;
        }

        void set(std::string _name, std::string _value)
        {
            int i = index(_name);

            if(i != -1)
            {
            	opts[i].set(_value);
                //adjust the column width if necessary
                col_width = (std::max)(col_width, opts[i].col_width());
            }
            else
                std::cout<<"ERROR - option not recognized: "<<_name<<std::endl;
        }

        //parse a parameter string
        void parse(int argc, char* argv[])
        {
            //if the number of options is 1, we're done
            if(argc <= 1) return;

            std::string name;
            std::string params;

            bool args_done = false;		//create a flag that turns true when the first option is encountered

            for(int i=1; i<argc; i++)
            {
                //if the argument is an option
                if(argv[i][0] == '-' && argv[i][1] == '-')
                {
                	args_done = true;				//arguments for the executable are done, all options now
                    //add any previous options
                    if(name != "")
                        set(name, params);
                    //set the current option to this name
                    name = argv[i]+2;
                    //clear the parameters list
                    params = "";
                }
                else if(!args_done){
                	args.push_back(argv[i]);
                }
                else{	//everything else is an arg for the most recent option
					if(params != "")
						params += " ";
                    params += argv[i];
                }
            }

            //set the last option
            set(name, params);
        }

        //determine if a parameter has been set (either specified by the user or with a default value)
        bool operator()(std::string _name)
        {
            int i = find(opts.begin(), opts.end(), _name) - opts.begin();

            if(i < 0)
            {
                std::cout<<"ERROR - Unspecified parameter name: "<<_name<<std::endl;
                exit(1);
            }

            return opts[i].is_set();
        }

        //number of arguments in a specified option
        unsigned int nargs(std::string _name)
        {
            int i = find(opts.begin(), opts.end(), _name) - opts.begin();

            if(i < 0)
            {
                std::cout<<"ERROR - Unspecified parameter name: "<<_name<<std::endl;
                exit(1);
            }

            return opts[i].nargs();
        }

        //number of arguments for the executable
        unsigned int nargs(){
        	return args.size();
        }

        //return the a'th executable argument
        std::string arg(unsigned int a){
        	return args[a];
        }

        cmd_option operator[](std::string _name)
        {
            int i = find(opts.begin(), opts.end(), _name) - opts.begin();

            if(i < 0 || i >= opts.size())
            {
                std::cout<<"ERROR - Unspecified parameter name: "<<_name<<std::endl;
                exit(1);
            }

            return opts[i];
        }


	};



}	//end namespace stim

#endif
